/* crypto/req.c */

#include <openssl/bio.h>
#include <openssl/err.h>
#include <openssl/rsa.h>
#include <openssl/evp.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/objects.h>
#include <openssl/pem.h>
#include <openssl/asn1.h>
#include <openssl/txt_db.h>

#include <ocsp_req_st.h>

#undef BSIZE
#define BSIZE 256

/* Allocation for an OCSP_REQINFO */
OCSP_REQINFO *OCSP_REQINFO_new() {
	OCSP_REQINFO *ret = NULL;

	ret = (OCSP_REQINFO *) OPENSSL_malloc (sizeof(OCSP_REQINFO));
	if( ret ) {
		/* If allocation went well, we want to erase
		   allocked memory */
		ret->version = NULL;
		ret->requestorName = NULL;
		ret->requestList = NULL;
		ret->requestExtensions = NULL;
	}

	return ret;
}

/* Allocation for an OCSP_CERTID structure */
OCSP_CERTID *OCSP_CERTID_new()
        {
        ASN1_CTX c;
        OCSP_CERTID *ret=NULL;

	if(!(ret = (OCSP_CERTID *) OPENSSL_malloc (sizeof(OCSP_CERTID))))
	{
		BIO_printf( bio_err, "ERROR: malloc CERTID_new\n");
		return(NULL);
	}

	/*
	ret->hashAlgorithm = X509_ALGOR_new();
	ret->issuerNameHash = ASN1_OBJECT_new();
	ret->issuerKeyHash = ASN1_OBJECT_new();
	ret->serialNumber = ASN1_OBJECT_new();
	*/

        M_ASN1_New(ret->hashAlgorithm, X509_ALGOR_new);
        M_ASN1_New(ret->issuerNameHash, ASN1_OCTET_STRING_new);
        M_ASN1_New(ret->issuerKeyHash, ASN1_OCTET_STRING_new);
        M_ASN1_New(ret->serialNumber, ASN1_INTEGER_new);

        return(ret);
	}

/* Allocation for an OCSP_SIGNATURE */
OCSP_SIGNATURE *OCSP_SIGNATURE_new(void) {
	OCSP_SIGNATURE *ret;

	if((ret = OPENSSL_malloc( sizeof( OCSP_SIGNATURE ))) == NULL )
		goto err;

	if((ret->signatureAlgorithm = X509_ALGOR_new()) == NULL ) {
		goto err;
	}

	if ((ret->signature = (ASN1_BIT_STRING*) ASN1_OBJECT_new()) == NULL) {
		goto err;
	}

	ret->certs 	= NULL;
	
	return (ret);

err:
	return(NULL);
}

/* Let's free an OCSP_SIGNATURE structure */
int OCSP_SIGNATURE_free( OCSP_SIGNATURE *sig ) {

	/* To be ended, it lacks signature_free and sig_alg_free */
	if( sig->certs )
		sk_pop_free( sig->certs, (void(*)()) X509_free);
	sig->certs = NULL;

	return(1);
}


/* Generate signature for the OCSP Request */
int OCSP_request_sign(OCSP_REQUEST *req, EVP_PKEY *key, EVP_MD *dgst,
		      STACK_OF(X509) *certs) {
	int i;
	OCSP_SIGNATURE *sig;

	/* Get an allocked signature structure */
	if (!(req->optionalSignature = sig = OCSP_SIGNATURE_new()))
		goto err;

	/* Let's sign the req->tbsRequest bytes */
	if( ! ASN1_sign((int(*)())i2d_OCSP_REQINFO,
			req->optionalSignature->signatureAlgorithm,
			NULL, sig->signature,
			(char *) req->tbsRequest,key,dgst))
		goto err;

	/* If there are certs, let's add them to the request,
	 * usually them will be requestor's and ca's certs */
	if (certs)
	        {
	        if (!(sig->certs = sk_dup(certs)))
			goto err;
	        for (i = 0; i < sk_num(sig->certs); i++)
	                {
		        if (!(sig->certs->data[i] = (char*)
		               X509_dup((X509*)certs->data[i]))) goto err;
		        }
		}
	return 1;
err:
	if (req->optionalSignature)
	        {
		OCSP_SIGNATURE_free(req->optionalSignature);
		req->optionalSignature = NULL;
		}
	return 0;
	}



OCSP_CERTID *OCSP_cert_id_new(EVP_MD *dgst, X509_NAME *issuerName, 
		ASN1_BIT_STRING* issuerKey, ASN1_INTEGER *serialNumber) {
        int 		nid;
        unsigned int 	i;
        X509_ALGOR 	*alg;
        OCSP_CERTID 	*cid = NULL;
        unsigned char 	md[EVP_MAX_MD_SIZE];

	/* Allocate the memory for the CID structure */
        if (!(cid = (OCSP_CERTID *) OPENSSL_malloc (sizeof(OCSP_CERTID)) ))
		goto err;

	/* Get the X509 Algorithm used in the certificate */
	if((alg = X509_ALGOR_new()) == NULL ) {
		goto err;
	}

	/* Get the digest algorithm used */
        if((nid = EVP_MD_type(dgst)) == NID_undef ) {
                goto err;
	}

	/* Set the CID hash algorithm */
	cid->hashAlgorithm = alg;

	/* If cannot set the algorithm, then error */
        if (!(alg->algorithm=OBJ_nid2obj(nid)))
		goto err;
        if ((alg->parameter=ASN1_TYPE_new()) == NULL)
		goto err;
        alg->parameter->type=V_ASN1_NULL;

        if (!X509_NAME_digest(issuerName, dgst, md, &i)) goto digerr;
        if (!(cid->issuerNameHash = ASN1_OCTET_STRING_new())) goto err;
        if (!(ASN1_OCTET_STRING_set(cid->issuerNameHash, md, i))) goto err;


	if( ! ASN1_digest((int (*)())i2d_ASN1_BIT_STRING,
				dgst,(char *)issuerKey,md,&i))
			goto digerr;

        if (!(cid->issuerKeyHash = ASN1_OCTET_STRING_new())) goto err;
        if (!(ASN1_OCTET_STRING_set(cid->issuerKeyHash, md, i))) goto err;
        
	/*
        if (cid->serialNumber != NULL) ASN1_INTEGER_free(cid->serialNumber);
	*/

	if (!(cid->serialNumber = ASN1_INTEGER_dup(serialNumber))) goto err;

        return cid;
digerr:
        /* OCSPerr(OCSP_F_CERT_ID_NEW,OCSP_R_DIGEST_ERR); */
	BIO_printf(bio_err,"ERROR: hash algorithm not supported\n");
err:
        /* if (cid) OCSP_CERTID_free(cid); */
	BIO_printf(bio_err,"ERROR: creating cert id structure\n");
        return NULL;
        }


int OCSP_request_add(OCSP_REQUEST *req, OCSP_CERTID *cid, STACK *extensions)
        {
        OCSP_ONEREQ *one = NULL;

        if (!(one = (OCSP_ONEREQ *) OPENSSL_malloc(sizeof(OCSP_ONEREQ))))
		goto err;

/*
	one->reqCert = (OCSP_CERTID *)ASN1_dup((int(*)())i2d_OCSP_CERTID,
			 (char *(*)())d2i_OCSP_CERTID,(char *)(cid));
*/
	
	one->reqCert = cid;

        if (!(one->reqCert))
		goto err;
/*
        if (extensions&&(!(one->singleRequestExtensions=ext_dup(extensions))))
                goto err;
*/
        if (!sk_push(req->tbsRequest->requestList,(char*)one)) {
		BIO_printf( bio_err, "ERROR: pushing request to requestList\n");
		goto err;
	}

        return 1;
err:
        return 0;
}

/* Printing Functions */
int OCSP_REQUEST_print(BIO *bp, OCSP_REQUEST *o) {
        int i,j,n;
        long l;
        char *s;
        OCSP_CERTID* cid = NULL;
        ASN1_INTEGER *ai = NULL;
        OCSP_ONEREQ *one = NULL;
        OCSP_REQINFO *inf = o->tbsRequest;
        OCSP_SIGNATURE *sig = o->optionalSignature;

        if (BIO_write(bp,"OCSP Request Data:\n",19) <= 0)
		goto err;

        l=ASN1_INTEGER_get(inf->version);
        if (BIO_printf(bp,"%4sVersion: %lu (0x%lx)","",l+1,l) <= 0)
		goto err;

        if (inf->requestorName != NULL)
        {
                if (BIO_write(bp,"\n    Requestor Name: ",21) <= 0) 
                        goto err;
                GENERAL_NAME_print(bp, inf->requestorName);
        }

        if (BIO_write(bp,"\n    Requestor List:\n",21) <= 0)
		goto err;

        for (i = 0; i < sk_num(inf->requestList); i++)
        {
                if (!inf->requestList->data[i])
			continue;
                one = (OCSP_ONEREQ*) inf->requestList->data[i];
                cid = one->reqCert;
                j=OBJ_obj2nid(cid->hashAlgorithm->algorithm);
                if (BIO_write(bp,"        Serial Number: ",23) <= 0) 
                        goto err;
                if (!i2a_ASN1_INTEGER(bp, cid->serialNumber)) 
                        goto err;
                if (BIO_printf(bp,"\n%12sHash Algorithm: %s","",
                               (j == NID_undef)?"UNKNOWN":OBJ_nid2ln(j)) <= 0)
                        goto err;
                if (BIO_write(bp,"\n            Issuer Name Hash: ",27) <= 0)
                        goto err;
                i2a_ASN1_STRING(bp, cid->issuerNameHash, V_ASN1_OCTET_STRING);
                if (BIO_write(bp,"\n            Issuer Key Hash: ",26) <= 0) 
                        goto err;
                i2a_ASN1_STRING(bp, cid->issuerKeyHash, V_ASN1_OCTET_STRING);
                if (!BIO_write(bp,"\n",1)) goto err;
		/*
                if (!OCSP_extensions_print(bp, one->singleRequestExtensions,
                                           "Request Single Extensions"))
                        goto err;
		*/
        }
/*
        if (!OCSP_extensions_print(bp, inf->requestExtensions, 
                                   "Request Extensions"))
                goto err;
*/
        if (sig)
                {

                i=OBJ_obj2nid(sig->signatureAlgorithm->algorithm);
                if (BIO_printf(bp,"    Signature Algorithm: %s",
                               (i == NID_undef)?"UNKNOWN":OBJ_nid2ln(i)) <= 0)
                        goto err;
		
                n=sig->signature->length;
                s=(char *)sig->signature->data;
                for (i=0; i<n; i++)
                        {
                        if ((i%18) == 0)
                        if (BIO_write(bp,"\n        ",9) <= 0) goto err;
                        if (BIO_printf(bp,"%02x%s",(unsigned char)s[i],
                                ((i+1) == n)?"":":") <= 0) goto err;
                        }
                if (BIO_write(bp,"\n",1) != 1) goto err;
                if (sig->certs)
                        {
                        for (i=0; i<sig->certs->num; i++)
                                if (sig->certs->data[i] != NULL)
                                        X509_print(bp,
                                                   (X509*)sig->certs->data[i]);
                        }
                }
        return 1;
err:
        return 0;
        }


int GENERAL_NAME_print(BIO *bp, GENERAL_NAME *n)
        {
        int j;
        char *p;

        if (n == NULL) return(0);

        switch (n->type)
                {

        case GEN_DIRNAME:
                X509_NAME_print(bp,n->d.dirn,16);
                break;

        case GEN_EMAIL:
        case GEN_DNS:
        case GEN_URI:
        case GEN_IPADD:
                p=(char *)n->d.ip->data;
                for (j=n->d.ip->length;j>0;j--)
                        {
                        if ((*p >= ' ') && (*p <= '~'))
                                BIO_printf(bp,"%c",*p);
                        else if (*p & 0x80)
                                BIO_printf(bp,"\\0x%02X",*p);
                        else if ((unsigned char)*p == 0xf7)
                                BIO_printf(bp,"^?");
                        else    BIO_printf(bp,"^%c",*p+'@');
                        p++;
                        }
                break;

        case GEN_RID:
                i2a_ASN1_OBJECT(bp, n->d.rid);
                break;

        /* XXX these are legit, need to support at some time... */
        case GEN_OTHERNAME:
        case GEN_X400:
        case GEN_EDIPARTY:
        default:
                return 0;
                }

        return 1;
        }

